const fs = require('fs')
const path = require('path')

const AdmZip = require('adm-zip')
const { write } = require('./files.js')
const config = require('./config.js')
const log = require('./log.js')

module.exports = {
  list: () => {
    const toReturn = []
    const folders = config.get('folders')
    const appFolders = fs.readdirSync(folders.applications)
    appFolders.forEach(appFolder => {
      try {
        const appDefinition = JSON.parse( fs.readFileSync(path.join(folders.applications,appFolder,'basics.json'), { encoding: 'utf8'}, true) )
        if (appDefinition.settings) toReturn.push(appDefinition)
      } catch(e) {}
    })
    return toReturn
  },
  load: (app) => {
    const appFolderDefinition = config.get('folders').applications
    const appFolders = fs.readdirSync(appFolderDefinition)
    appFolders.forEach(appFolder => {
      const basicsPath = path.join(appFolderDefinition, appFolder, 'basics.json')
      if (fs.existsSync(basicsPath)) { // it is an app
        try {
          const appDefinition = JSON.parse( fs.readFileSync(basicsPath, { encoding: 'utf8'}, true) )
          if ((app._id && appDefinition._id === app._id) || appDefinition._id === app) {
            if (typeof app === 'string') {
              app = { ...appDefinition }
            }
            app.folder = appFolder
            app.tables = JSON.parse( fs.readFileSync( path.join( appFolderDefinition, appFolder, 'tables.json' ), { encoding: 'utf8'}, true) )
            app.pages = JSON.parse( fs.readFileSync( path.join( appFolderDefinition, appFolder, 'pages.json' ), { encoding: 'utf8'}, true) )
            app.assets = JSON.parse( fs.readFileSync( path.join( appFolderDefinition, appFolder, 'assets.json' ), { encoding: 'utf8'}, true) )
          }
        } catch(e) {
          const theError = e.exitCode ? e : { exitCode: 1, message: 'Error parsing your Apps (backup maybe?)', error: e }
          throw(theError)
        }

        const makeItPlain = (element) => {
          return element.map(ele => {
            const elementPath = path.join( appFolderDefinition, appFolder, 'Pages', `${ele.unique_id}.json` )
            let eleDefinition
            if (fs.existsSync(elementPath)) {
              eleDefinition = JSON.parse( fs.readFileSync(elementPath, { encoding: 'utf8'}, true) )
              if (ele.children) ele.children = makeItPlain(ele.children)
              aptugocli.plain[ele.unique_id] = { ...ele, ...eleDefinition }
            } else {
              eleDefinition = { warning: `${ele.unique_id}.json does not exist`}
            }
            return { ...ele, ...eleDefinition }
          })
        }

        if (app.folder) {
          app.pages = makeItPlain(app.pages)

          app.tables = app.tables.map(table => {
            const tableDefinition = JSON.parse( fs.readFileSync( path.join( appFolderDefinition, app.folder, 'Tables', `${table.unique_id}.json` ), { encoding: 'utf8'}, true) )
            aptugocli.plainTables[table.unique_id] = tableDefinition
            tableDefinition.fields.forEach(field => {
              if (!aptugocli.plainFields) aptugocli.plainFields = {}
              aptugocli.plainFields[field.unique_id] = field
            })
            return tableDefinition
          })
        }
      }

    })
    return app
  },
  save: (app) => {
    const folders = config.get('folders')
    const appFolders = folders.applications
    const saveFolder = path.join(appFolders, aptugocli.friendly(app.settings.name))

    aptugocli.createIfDoesntExists(saveFolder)
    aptugocli.createIfDoesntExists(path.join(saveFolder,'Tables'))
    aptugocli.createIfDoesntExists(path.join(saveFolder,'Pages'))

    let { _id, createdAt, modifiedAt, lastBuild, ...rest } = app
    modifiedAt = new Date().getTime()
    write({ saveFolder: saveFolder, filename: 'basics.json', content: { _id, createdAt, modifiedAt, lastBuild, settings: { ...app.settings } } })
    if (app.tables) write({ saveFolder: saveFolder, filename: 'tables.json', content: app.tables, clean: 'Tables' })
    if (app.pages) write({ saveFolder: saveFolder, filename: 'pages.json', content: app.pages, clean: 'Pages' })
    if (app.assets) write({ saveFolder: saveFolder, filename: 'assets.json', content: app.assets })
    return app
  },
  remove: (app) => {
    const folders = config.get('folders')
    const appFolders = folders.applications
    const saveFolder = path.join(appFolders, aptugocli.friendly(app.settings.name))
    fs.rmdirSync(saveFolder, { recursive: true })
  },
  createZip: (app, options = { what: 'source', method: 'development' }) => {

      if (!options.what) options.what = 'app'
      if (!options.method) options.method = 'development'

      const settings = app.settings[options.method]
      const buildFolder = settings.folder
      let fullbuildfolder = config.get('folders').applications

      if (options.what === 'source') {
        fullbuildfolder = config.get('folders').build
      }
      const sourceFolder = path.join(fullbuildfolder, buildFolder)
      const zipName = `${aptugocli.friendly(app.settings.name)}_${options.what}_${options.method}.zip`

      var zip = new AdmZip()
      let dirFiles = fs.readdirSync(sourceFolder)
      dirFiles.forEach((file) => {
        if(fs.lstatSync(path.join(sourceFolder, file)).isDirectory()) {
          log(`Adding directory: ${file}`, { verbosity: 9 })
          zip.addLocalFolder(path.join(sourceFolder, file), file)
        } else {
          log(`Adding file: ${file}`, { verbosity: 10 })
          zip.addLocalFile(path.join(sourceFolder, file))
        }
      })
      log(`Zip file written`, { verbosity: 7 })
      zip.writeZip(zipName)
      return path.join(process.cwd(), zipName)
  }
}
