const path = require('path')
const os = require('os')
const util = require('util')
const fs = require('fs')
const { Blob } = require('buffer')

const twigPkg = require('twig/twig.js')
const { extendFilter, extendFunction, twig: _twig, cache } = twigPkg

const apps = require('../apps.js')
const templates = require('../templates.js')

const { stateSync: loadState, setState: updateState } = require('../state.js')
const config = require('../config.js')
const copyAssets = require('./copyassets.js')
const loadElements = require('./loadElements.js')
const te = require('./twigExtensions.js')
const saveTwigTemplates = require('./twigTemplates.js')
const copyStaticFiles = require('./copystaticfiles.js')
const copyExtraFiles = require('./copyextrafiles.js')
const buildPage  = require('./buildpage.js')
const log = require('../log.js')
const ora = require('ora')
const chalk = require('chalk')
const humanizeDuration = require('humanize-duration')
const parseToString = require('./parseToString')
const {spawn, execSync, exec} = require('child_process')
const AdmZip = require('adm-zip')
const FormData = require('form-data')
const error = require('../error.js')
const errored = false
const { randomFillSync } = require('crypto')
const axios = require('axios')
const fmdata = require('formdata-node')
const fixPath = require('./paths/fixPath.js')

let child
const random = (() => {
  const buf = Buffer.alloc(16)
  return () => randomFillSync(buf).toString('hex')
})()

const twigExtensions = () => {
  extendFunction('parse', te.parse)
  extendFunction('includeTemplate', te.includeTemplate)
  extendFunction('insert_setting', te.insertSetting)
  extendFunction('save_delayed', te.saveDelayed)
  extendFunction('add_setting', te.addSetting)
  extendFunction('envVar', te.envVar)

  extendFilter('elementData', te.elementData)
  extendFilter('fieldData', te.fieldData)
  extendFilter('removeExtension', te.removeExtension)
  extendFilter('plain', te.plain)
  extendFilter('tableData', te.tableData)
  extendFilter('assetData', te.assetData)
  extendFilter('castToArray', te.castToArray)

  extendFilter('camelCase', function(value, params) {
    return value.replace(/(?:^\w|[A-Z]|\b\w|\s+)/g, function(match, index) {
      if (+match === 0) return ""; // or if (/\s+/.test(match)) for white spaces
      return index === 0 ? match.toLowerCase() : match.toUpperCase();
    })
  })

  extendFilter('friendly', function(value, params) {
    return value ? value.replace(/[^a-zA-Z0-9]/g, "") : ''
  })

  extendFilter('functionOrCall', function(value) {
    if (!value) return ''
    if (/^\([a-zA-Z ,]*\) => {/.exec(value)) return value
    if (/^[a-zA-Z0-9]*$/.exec(value)) return value
    return '(e) => {\n' + value + '\n}'
  })
  extendFilter('textOrVariable', function(value) {
    if (!value) return ''
    if (typeof value === "number") return `"${value}"`
    if (/\${.*}/.exec(value)) return "{`" + value + "`}"
    else if (value.substr(0,1) === '{') return value
    else if (value.substr(0,4) === 'var:') return value.substr(4)
    else return `"${value}"`
  })

  extendFilter('withoutVars', te.withoutVars)

  // extendFilter('parentPage', function(value) {
  //   const parentsTree = getCascadingTree(builderObj.plainPages, value.unique_id)
  //   return parentsTree
  // })

}

const builder = {
  build: async (prefs) => {
    const { app, type = 'Development', clean = false, skip = [], only = null } = prefs

    if (!only) log(`Building ${app.settings.name} in ${type} mode`, { type: 'mainTitle' })
    let parameters
    try {
      parameters = builder.buildParameters({ app, type, clean, variables: {}, skip })
    } catch(e) {
      throw(e)
    }

    if (only) {
      switch (only) {
        case 'stop':
          return builder.stopBuilding()
          break
        case 'setup':
          return builder.firstStep_setupBuild(parameters).then(() => {
            return 'finished setup'
          }).catch(e => {
            console.log('caught Error on SetupBuild', e)
          })
          break
        case 'check':
          return builder.secondStep_checkApplication(parameters).then((response) => {
            return 'finished check'
          }).catch(e => {
            // const theError = { exitCode: 112, message: 'Ran into an issue checking your application'}
            // console.warn(theError)
            return e
          })
        case 'copy':
          return builder.thirdStep_copyStaticFiles(parameters).then((res) => {
            return 'finished copy'
          }).catch(e => {
            console.log('Caught Error on Copy Static Files', e)
          })
          break
        case 'pages':
          if (parameters.stoped) return
          return builder.fourthStep_generatePages(parameters).then(() => {
            return 'finished pages'
          }).catch(e => {
            return e
          })
          break
        case 'extraFiles':
          if (parameters.stoped) return
          return builder.fifthStep_extraFiles(parameters).then(() => {
            return 'finished extrafiles'
          })
        case 'extra':
          if (parameters.stoped) return
          return builder.fifthStep_extraSettings(parameters).then(() => {
            return 'finished extra'
          })
          break
        case 'post':
          return builder.sixthStep_postBuild(parameters).then(() => {
            return 'finished post'
          })
          break
        case 'buildscripts':
          return builder.seventhStep_buildScripts(parameters).then(() => {
            return 'finished buildscripts'
          })
          break
        case 'deploy':
          return builder.eightStep_deploy(parameters).then(() => {
            return 'finished deployment'
          }).catch(e => {
            console.log('step 8: Error', e)
            throw(e)
          })
          break
      }
      return 'error' + only
    } else {

      let result = await builder.firstStep_setupBuild(parameters)
      if (result?.exitCode) return result

      result = await builder.secondStep_checkApplication(parameters)
      if (result?.exitCode) return result

      result = await builder.thirdStep_copyStaticFiles(parameters)
      if (result?.exitCode) return result

      try {
        result = await builder.fourthStep_generatePages(parameters)
        if (result?.exitCode) return result
      } catch(e) {
        return e
      }

      result = await builder.fifthStep_extraFiles(parameters)
      if (result?.exitCode) return result

      result = await builder.fifthStep_extraSettings(parameters)
      if (result?.exitCode) return result

      result = await builder.sixthStep_postBuild(parameters)
      if (result?.exitCode) return result

      result = await builder.seventhStep_buildScripts(parameters)
      if (result?.exitCode) return result

      result = await builder.lastStep_success(parameters)
      if (result?.exitCode) return result

      // builder.firstStep_setupBuild(parameters).then(() => {
      //   if (typeof aptugo !== 'undefined') aptugo.setFeedback('Build Setup...')
      //   builder.secondStep_checkApplication(parameters).then(() => {
      //     if (typeof aptugo !== 'undefined') aptugo.setFeedback('Check your Application...')
      //     builder.thirdStep_copyStaticFiles(parameters).then(() => {
      //       if (typeof aptugo !== 'undefined') aptugo.setFeedback('Copy Static Files...')
      //       builder.fourthStep_generatePages(parameters).then(() => {
      //         if (typeof aptugo !== 'undefined') aptugo.setFeedback('Generate Pages...')
      //         builder.fifthStep_extraFiles(parameters).then(() => {
      //           if (typeof aptugo !== 'undefined') aptugo.setFeedback('Copy Files importd by Elements...')
      //           builder.fifthStep_extraSettings(parameters).then(() => {
      //             if (typeof aptugo !== 'undefined') aptugo.setFeedback('Rebuild Pages with extra settings...')
      //             builder.sixthStep_postBuild(parameters).then(() => {
      //               if (typeof aptugo !== 'undefined') aptugo.setFeedback('Post build stuff...')
      //               builder.seventhStep_buildScripts(parameters).then(() => {
      //                 if (typeof aptugo !== 'undefined') aptugo.setFeedback('Post build scripts...')
      //                 builder.lastStep_success(parameters).then(() => {
      //                   if (typeof aptugo !== 'undefined') aptugo.setFeedback('done')
      //                   return 'Built Full'
      //                   // finished
      //                 })
      //               })
      //             })
      //           })
      //         })
      //       }).catch(e => { // Failed generate pages
      //         console.log('here', e)
      //         return e
      //       })
      //     })
      //   })
      // })
      
    }
  },

  stopBuilding: (parameters) => {
    errored = true
    parameters.stoped = true
    child.stdin.pause()
    child.kill()
  },

  parseApplication: (application) => {
    try {
      application.tables.forEach(table => aptugocli.plain[table.unique_id] = table)
      application.assets.forEach(asset => {
        aptugocli.plain[asset.id] = asset
        aptugocli.plainAssets[asset.id] = asset
      })

      const navigateAndParseTree = (tree) => {
        return tree.map(item => {
          aptugocli.plain[item.unique_id] = item
          if (item.type === 'page') {
            // aptugocli.plain[page.unique_id] = item
            // aptugocli.plainPages[page.unique_id] = item
            Object.keys(item).map(propertyName => {
              if (item[propertyName] && item[propertyName].substr && item[propertyName].substr(0,2) === '()') {
                let replacedValue = item[propertyName].replace('aptugo.store.getState().application.tables','params.plainTables')
                replacedValue = replacedValue.replace('aptugo.activeApplication.tables','params.plainTables')
                replacedValue = '(params)' + replacedValue.substr(2)
                item[propertyName] = parseToString(replacedValue)
              }
            })
          }
          if (item.children && item.children.length) item.children = navigateAndParseTree(item.children)
          return item
        })
      }
      application.pages = navigateAndParseTree(application.pages)
      return application
    } catch(e) {
      throw(e)
    }
  },

  buildParameters: (buildData) => {
    try {
      const application = builder.parseApplication(apps.load(buildData.app))

      // console.log(util.inspect(application, false, null, true /* enable colors */))
      const settings = buildData.type === 'Development' ? application.settings.development : application.settings.production
      const template = templates.get(settings.template)
      const buildFolder = settings.folder
      const fullbuildfolder = config.get('folders').build
      const appFolder =  path.join( config.get('folders').applications, aptugocli.friendly(application.settings.name) )
      const filesFolder = path.join(appFolder, 'Drops')

      settings.variables && settings.variables.split('\n').forEach(thevar => {
        var [varName, varValue] = thevar.split(':')
        try { buildData.variables[varName] = eval(varValue) } catch(e) { buildData.variables[varName] = varValue }
      })

      if (!template) error('Error: Application does not have a template assigned (or it is missing)', true)

      aptugocli.activeParameters = {
        stoped: false,
        skip: buildData.skip,
        type: buildData.type,
        application,
        settings,
        template,
        buildFolder,
        appFolder,
        fullbuildfolder,
        filesFolder,
        method: settings.type,
        doClean: buildData.clean,
        variables: buildData.variables
      }
      return aptugocli.activeParameters
    } catch(e) {
      throw(e)
    }
  },

  firstStep_setupBuild: (parameters) => {
    return new Promise((resolve, reject) => {
      try {
        const state = loadState()
        const start = new Date()
        console.info('Setting up build...')
        updateState({ ...state, filesWithExtraSettings: [], extraSettings: {} })
        if (parameters.doClean) {
          fs.rmdirSync( path.join(parameters.fullbuildfolder, parameters.buildFolder), { recursive: true })
        }
        aptugocli.createIfDoesntExists(path.join(parameters.fullbuildfolder, parameters.buildFolder))
  
        aptugocli.skipSettings = true
        aptugocli.filesimportdByElements = []
        cache(false)
        twigExtensions()
        _twig({
          id: 'empty',
          allowInlineIncludes: true,
          data: '',
          rethrow: true
        })
  
        const elementsFolder = parameters.template.files.filter(file => file.path === 'elements' || file.path === 'templatescripts')[0]
        if (!elementsFolder) error('ERROR IN TEMPLATE: Template does not contain an elements folder', false)
        else {
          aptugocli.loadedElements = loadElements(elementsFolder.children)
          saveTwigTemplates(elementsFolder.children)
        }
        aptugocli.assets = parameters.application.assets
        const end = new Date()
        console.info(`Build set-up: ${humanizeDuration(end - start)}`, { finished: true })
        resolve()
      } catch(e) {
        reject(e)
      }
    })
  },

  secondStep_checkApplication: (parameters) => {
    return new Promise((resolve, reject) => {
      const start = new Date()
      console.info('Checking your application...')
      try {
        // Check Tables
        const tables = parameters.application.tables
        tables.forEach(table => {
          if ( aptugocli.friendly(table.name) === aptugocli.friendly(table.singleName) ) {
            throw { element: table, error: 'tablenamesame', type: 'table' }
          }

          table.fields.forEach(field => {
            const definition = parameters.template.fields.find(tplfield => tplfield.value === field.data_type)
            if (!definition) {

              throw { element: field, error: 'nodefinition', type: 'field' }
            }

            definition.options && definition.options.extra && definition.options.extra.forEach(extraDefinition => {
              if (extraDefinition.isimportd && !field[extraDefinition.value]) throw { element: field, error: 'missingimportd', type: 'field' }
            })
          })
        })

        // Check Elements
        const elements = Object.values(aptugocli.plain).filter(plainThing => plainThing.type === 'element')
        elements.forEach(element => {
          const broughtElement = aptugocli.loadedElements.find(item => item.path === `${element.value}.tpl`)
          if (broughtElement) {
            broughtElement.options && broughtElement.options.forEach(option => {
              if (option.importd && !element.values[option.name]) throw { element: element, error: 'missingimportd', type: 'element' }
            })
          } else {
            throw { element: element, error: 'missingtemplate', type: 'element' }
          }

        })
        const end = new Date()
        console.info(`Finished checking your application: ${humanizeDuration(end - start)}`, { finished: true })
        resolve()
      } catch(e) {
        const theError = { exitCode: 124, message: 'Something fishy', element: e.element, type: e.type || 'element' }
        if (e.error === 'tablenamesame') theError.message = `${e.element.name} can not have the same name for Single Values`
        else if (e.error === 'missingimportd') {
          if (e.type === 'field') theError.message = `${e.element.column_name} has unfilled importd definitions`
          else theError.message = `${e.element.name} has unfilled importd definitions`
        } else if (e.error === 'nodefinition') theError.message = `Your template does not support fields of type: ${e.element.column_name}`
        else if (e.error === 'missingtemplate') theError.message = `Your template does not support elements of type: ${e.element.value}`
        reject(theError)
      }
    })
  },

  thirdStep_copyStaticFiles: (parameters) => {
    return new Promise((resolve, reject) => {
      const start = new Date()
      console.info('Copying static files...')

      if (parameters.skip.indexOf('copy') === -1) {
        try {
          copyAssets(parameters)
          copyStaticFiles(parameters)
          const end = new Date()
          console.info(`Static files copied: ${humanizeDuration(end - start)}`, { finished: true })
          resolve()
        } catch(e) {
          const theError = {
            exitCode: 120,
            message: 'Missing Asset File',
            error: e
          }
          console.warn(theError)
          reject(theError)
        }
      } else {
        const end = new Date()
        console.info(`Static files skiped: ${humanizeDuration(end - start)}`, { finished: true })
        resolve()
      }
    })
  },

  fourthStep_generatePages: (parameters) => {
    return new Promise((resolve, reject) => {
      if (parameters.skip.indexOf('pages') !== -1) {
        console.info(`Pages generation skiped`)
      } else {
        const start = new Date()
        console.info('Generating pages...')
        aptugocli.generationFolder = parameters.template.renderingFolder ? path.join(parameters.buildFolder, parameters.template.renderingFolder ) : templates.searchForRenderingPlaceholder(parameters.template.files, parameters.buildFolder)
        try {
          parameters.application.pages.forEach(page => {
            let returned = buildPage(page, parameters)
            if (returned.exitCode) {
              throw(returned)
            }
          })
          const end = new Date()
          console.info(`Pages generated: ${humanizeDuration(end - start)}`, { finished: true })
          resolve()
        } catch(e) {
          const theError = e.exitCode ? e : { exitCode: 123, message: 'Rude, condescending, and a little bit snotty error', element: page, type: 'page', error: e }
          reject(theError)
        }
      }
    })
  },

  fifthStep_extraFiles: (parameters) => {
    return new Promise((resolve, reject) => {
      const start = new Date()
      console.info('Copying extra files from elements...')
      if (parameters.skip.indexOf('extraFiles') === -1) {
        copyExtraFiles({ ...parameters, files: aptugocli.filesimportdByElements })
        const end = new Date()
        console.info(`Extra Files copied: ${humanizeDuration(end - start)}`, { finished: true })
      } else {
        const end = new Date()
        console.info(`Extra Files skiped: ${humanizeDuration(end - start)}`, { finished: true })
      }
      resolve()
    })
  },

  fifthStep_extraSettings: (parameters) => {
    return new Promise((resolve, reject) => {
      const state = loadState()
      const start = new Date()
      state.extraSettings = { ...state.extraSettings, ...parameters.variables }
      updateState(state)
      console.info('Re-Generating pages with extra settings...')
      
      if (parameters.skip.indexOf('copy') === -1) {
        aptugocli.skipSettings = false
        copyStaticFiles({ ...parameters, files: state.filesWithExtraSettings })
        const end = new Date()
        console.info(`Static files re-generated: ${humanizeDuration(end - start)}`, { finished: true })
      } else {
        const end = new Date()
        console.info(`Static files skiped (2nd pass): ${humanizeDuration(end - start)}`, { finished: true })
      }
      resolve()
    })
  },

  sixthStep_postBuild: (parameters) => {
    return new Promise((resolve, reject) => {
      const start = new Date()
      console.info('Running post-build commands...')
      if (parameters.skip.indexOf('post') === -1) {
        const tca = parameters.type === 'Development' ? parameters.template.templateCommandsAfter : parameters.template.templateCommandsAfterDeploy
        if (tca) {
          var template = _twig({ data: tca, rethrow: true })
          const renderedCommands = template.render(parameters)
          const baseFilesFolder = path.join(parameters.fullbuildfolder, parameters.buildFolder)
          const child = spawn(`cd ${baseFilesFolder} && ${renderedCommands}`, {
            shell: true
          })

          child.stderr.on('data', function (data) {
            console.log(data.toString())
          })

          child.stdout.on('data', function (data) {
            log(data.toString(), { verbosity: 8 })
          })

          child.on('exit', function (exitCode) {
            if (exitCode > 0) {
              console.log(`Child exited with code: ${exitCode}`, exitCode)
              // reject({ message: 'Postbuild Finished with code: ' + exitCode, error: exitCode })
            } else {
              const end = new Date()
              console.info(`Post build comands finished: ${humanizeDuration(end - start)}`, { finished: true });
              child.stdin.pause()
              child.kill()
              resolve()
            }
          })
        } else {
          const end = new Date()
          console.info(`No post build commands to run: ${humanizeDuration(end - start)}`, { finished: true })
          resolve()
        }
      } else {
        const end = new Date()
        console.info(`Post build commands skiped: ${humanizeDuration(end - start)}`, { finished: true })
        resolve()
      }
    })
  },

  seventhStep_buildScripts: (parameters) => {
    return new Promise((resolve, reject) => {
      const start = new Date()
      console.info('Running post-build scripts...')
      var isWin = process.platform === "win32"
      // FIX PATH
      let returnValue = {}

      // try {
      //   console.info('c')
      //   let result = exec(`${ os.userInfo().shell} -ilc 'echo -n "_SHELL_ENV_DELIMITER_"; env; echo -n "_SHELL_ENV_DELIMITER_"; exit'`, { shell: true })
      //   console.info('d')
      //   if (result) result = result.toString()
      //   // for (const line of result.split('\n').filter(line => Boolean(line))) {
      //   //   const [key, ...values] = line.split('=');
      //   //   returnValue[key] = values.join('=');
      //   // }
      // } catch(e) {
      //   console.info('e')
      //   console.log(e)
      // }

      if (returnValue.PATH) process.env.PATH = returnValue.PATH
      let precommand = ''

      Object.keys(parameters.variables).forEach(avar => {
        if (avar.substring(0,4) === 'ENV.') {
          precommand += `export ${avar.substring(4)}=${parameters.variables[avar]} &&`
        }
      })

      if (parameters.skip.indexOf('post') === -1) {
        let command = null
        const folders = config.get('folders')
        const settingsBuild = config.get('templates')

        if (parameters.type === 'Development') {
          try {
            command = settingsBuild[parameters.settings.template].build || 'pnpm i'
          } catch(e) {
            const scriptFolder = path.join(folders.templates, parameters.settings.template, 'templatescripts', isWin ? 'development.bat' : 'development.sh')
            if ( fs.existsSync( scriptFolder ) ) {
              command = scriptFolder
            }
          }
        } else {
          try {
            command = settingsBuild[parameters.settings.template].prodbuild || 'pnpm i && pnpm run build'
          } catch(e) {
            const scriptFolder = path.join(folders.templates, parameters.settings.template, 'templatescripts', isWin ? 'production.bat' : 'production.sh')
            if ( fs.existsSync( scriptFolder ) ) {
              command = scriptFolder
            }
          }
        }

        if (command) {
          fixPath().then(() => {
            const baseFilesFolder = path.join(parameters.fullbuildfolder, parameters.buildFolder)
            child = spawn(`${precommand} cd ${baseFilesFolder} && ${command}`, {
              shell: true,
              env: {
                PATH: process.env.PATH
              }
            })

            child.stderr.on('data', function (data) {
              log(data.toString(), { verbosity: 0 })
            })

            child.stdout.on('data', function (data) {
              log(data.toString(), { verbosity: 5 })
            })

            child.on('exit', function (exitCode) {
              if (exitCode > 0) {
                console.error(`Child exited with code: ${exitCode}`)
                reject({ message: 'Postbuild Finished with code: ' + exitCode, error: exitCode })
              } else {
                const end = new Date()
                console.info(`Post build scripts finished: ${humanizeDuration(end - start)}`, { finished: true });
                resolve()
              }
            })
          })
        } else {
          const end = new Date()
          console.info(`No post build scripts to run: ${humanizeDuration(end - start)}`, { finished: true });
          resolve()
        }
      } else {
        const end = new Date()
        console.info(`Post build scripts skiped: ${humanizeDuration(end - start)}`, { finished: true })
        resolve()
      }
    })
  },

  eightStep_deploy: (parameters) => {
    return new Promise((resolve, reject) => {
      const start = new Date()
      console.info('Starting Deployment...')

      // zip front-end
      let zipFilesFolder = path.join(parameters.fullbuildfolder, parameters.buildFolder, 'build')
      let saveTo = path.join(os.tmpdir(), 'aptugo', `aptugoapp-build-${random()}.zip`)

      var zip = new AdmZip()

      let dirFiles = fs.readdirSync(zipFilesFolder)
      dirFiles.forEach((file) => {
        if(fs.lstatSync(path.join(zipFilesFolder, file)).isDirectory()) {
          log(`Adding directory: ${file}`, { verbosity: 9 })
          zip.addLocalFolder(path.join(zipFilesFolder, file), file)
        } else {
          log(`Adding file: ${file}`, { verbosity: 10 })
          zip.addLocalFile(path.join(zipFilesFolder, file))
        }
      })
      log(`Zip file written`, { verbosity: 7 })
      zip.writeZip(saveTo)

      // post
      const url = parameters.settings.url.substring(8).toLowerCase()
      const buffer = zip.toBuffer()
      const form = new FormData()
      form.append('data', buffer)

      const options = {
        url: 'https://appuploader.aptugo.app:8501/upload?appName=' + url + '&user=' + aptugocli.user,
        method: 'POST',
          headers: { 'Content-Type': `multipart/form-data; boundary=${form._boundary}` },
        data: form
      }

      log(`Updloading front-end to Aptugo Servers`, { verbosity: 6 })
      axios(options)
        .then(response => {
          log('Front-end uploaded', { verbosity: 7 })
        })
        .catch(error => {
          console.warn('failed uploading', error)
        })

      // zip backend
      const zipFilesFolderbe = path.join(parameters.fullbuildfolder, parameters.buildFolder, 'back-end')
      saveTo = path.join(os.tmpdir(), 'aptugo', `aptugoapp-build-backend-${random()}.zip`)
      var zipbe = new AdmZip()

      dirFiles = fs.readdirSync(zipFilesFolderbe)
      dirFiles.forEach((file) => {
        if (file !== 'node_modules') {
          if(fs.lstatSync(path.join(zipFilesFolderbe, file)).isDirectory()) {
            console.log('...adding directory', file)
            zipbe.addLocalFolder(path.join(zipFilesFolderbe, file), file)
          } else {
            console.log('...adding file', file)
            zipbe.addLocalFile(path.join(zipFilesFolderbe, file))
          }
        }
      })

      zipbe.writeZip(saveTo)

      // post back-end
      const beurl = parameters.settings.apiURL.substring(8)
      const bebuffer = zipbe.toBuffer()
      const beform = new FormData()
      beform.append('data', bebuffer)

      const beoptions = {
        url: 'https://appuploader.aptugo.app:8501/upload?type=be&appName=' + beurl + '&user=' + aptugocli.user,
        method: 'POST',
          headers: { 'Content-Type': `multipart/form-data; boundary=${beform._boundary}` },
        data: beform
      }

      axios(beoptions)
        .then(response => {
          log('Back-end uploaded', { verbosity: 7 })
        })
        .catch(error => {
          console.warn('Error Uploading', error)

        })
      const end = new Date()
      console.info(`Deployment queued (might take a few minutes): ${humanizeDuration(end - start)}`, { finished: true })
      resolve()
    })
  },

  lastStep_success: (parameters) => {
    return new Promise((resolve, reject) => {

        process.stdout.write(chalk.hex('#FF603D')(`\n\r
        ┌────────────────────────┐
        │                        │
        │   Application Built!   │
        │                        │
        └────────────────────────┘
        Your application has been built on folder:
        ${path.join(parameters.fullbuildfolder, parameters.buildFolder)}
        `))
        process.exit(0)

      resolve()
    })
  }
}

module.exports = builder
