const execa = require('execa')
const { exec } = require('child_process')

function ansiRegex({onlyFirst = false} = {}) {
	const pattern = [
		'[\\u001B\\u009B][[\\]()#;?]*(?:(?:(?:(?:;[-a-zA-Z\\d\\/#&.:=?%@~_]+)*|[a-zA-Z\\d]+(?:;[-a-zA-Z\\d\\/#&.:=?%@~_]*)*)?\\u0007)',
		'(?:(?:\\d{1,4}(?:;\\d{0,4})*)?[\\dA-PR-TZcf-nq-uy=><~]))'
	].join('|');

	return new RegExp(pattern, onlyFirst ? undefined : 'g');
}

function stripAnsi(string) {
	if (typeof string !== 'string') {
		throw new TypeError(`Expected a \`string\`, got \`${typeof string}\``);
	}
	return string.replace(ansiRegex(), '');
}

const detectDefaultShell = () => {
	const {env} = process;

	if (process.platform === 'win32') {
		return env.COMSPEC || 'cmd.exe';
	}

	try {
		const {shell} = userInfo();
		if (shell) {
			return shell;
		}
	} catch {}

	if (process.platform === 'darwin') {
		return env.SHELL || '/bin/zsh';
	}

	return env.SHELL || '/bin/sh';
}

const defaultShell = detectDefaultShell();

const args = [
	'-ilc',
	'echo -n "_SHELL_ENV_DELIMITER_"; env; echo -n "_SHELL_ENV_DELIMITER_"; exit',
]

const env = {
	// Disables Oh My Zsh auto-update thing that can block the process.
	DISABLE_AUTO_UPDATE: 'true',
};

const parseEnv = env => {
	env = env.split('_SHELL_ENV_DELIMITER_')[1];
	const returnValue = {};

	for (const line of stripAnsi(env).split('\n').filter(line => Boolean(line))) {
		const [key, ...values] = line.split('=');
		returnValue[key] = values.join('=');
	}

	return returnValue;
};

async function shellEnv(shell) {
	if (process.platform === 'win32') {
		return process.env;
	}

	try {
		const {stdout} = await execa(shell || defaultShell, args, {env});
		return parseEnv(stdout);
	} catch (error) {
		if (shell) {
			throw error;
		} else {
			return process.env;
		}
	}
}

function shellEnvSync(shell) {

    if (process.platform === 'win32') {
      return process.env;
    }

    try {
      const {stdout} = execa.execaSync(shell || defaultShell, args, {env});
      return parseEnv(stdout);
    } catch (error) {
      if (shell) {
        throw error;
      } else {
        return process.env;
      }
    }
}

function shellPathSync() {
	const {PATH} = shellEnvSync();
	return PATH;
}

module.exports = () => {
  return new Promise((resolve, reject) => {
    if (process.platform === 'win32') resolve()
    let result
    exec(`${detectDefaultShell()} -ilc env`, { encoding: 'utf8', timeout: 3000 }, (error, stdout, stderr) => {
      if (error) {
        console.log(`exec error: ${error}`);
        resolve()
      }
      result = stdout

      const returnValue = {}
      for (const line of stripAnsi(result).split('\n').filter(line => Boolean(line))) {
        const [key, ...values] = line.split('=');
        returnValue[key] = values.join('=');
      }
      const {PATH} = returnValue
      process.env.PATH = PATH || ['./node_modules/.bin','/.nodebrew/current/bin','/usr/local/bin', process.env.PATH,].join(':');
      resolve()
    })
  })

}
